use super::*;

/// [`Mode`] for [`RegularFactory`] and [`RegularAtomic`].
///
/// [`RegularFactory`]: crate::rcore::RegularFactory
/// [`RegularAtomic`]: crate::atomic::RegularAtomic
pub struct RegularMode;

impl Mode for RegularMode {
    /// Discards the parser.
    type ParseSuccess<A, I> = A;

    /// Tries to extend the value.
    type ExtensionResult<A, E> = Result<A, E>;

    /// Keeps the value.
    type ExtensionSource<A> = A;

    fn bind<A0, A1, E, I>(
        s: Self::ParseSuccess<A0, I>,
        f: impl FnOnce(A0) -> Result<A1, E>,
    ) -> ModeResult<Self, A1, E, I> {
        f(s)
    }

    fn map<A0, A1, I>(
        s: Self::ParseSuccess<A0, I>,
        f: impl FnOnce(A0) -> A1,
    ) -> Self::ParseSuccess<A1, I> {
        f(s)
    }

    fn seal<A, I>(s: Self::ParseSuccess<A, I>) -> A {
        s
    }

    fn xmap_err<A, E0, E1>(
        result: Self::ExtensionResult<A, E0>,
        f: impl FnOnce(E0) -> E1,
    ) -> Self::ExtensionResult<A, E1> {
        result.map_err(f)
    }

    fn xbind<A0, A1, E>(
        result: Self::ExtensionResult<A0, E>,
        f: impl FnOnce(A0) -> Result<A1, E>,
    ) -> Self::ExtensionResult<A1, E> {
        result.and_then(f)
    }

    fn xseal<A, E>(result: Self::ExtensionResult<A, E>) -> Result<A, E> {
        result
    }

    fn smap<A0, A1>(
        source: Self::ExtensionSource<A0>,
        f: impl FnOnce(A0) -> A1,
    ) -> Self::ExtensionSource<A1> {
        f(source)
    }

    fn prepare<A>(a: A) -> Self::ExtensionSource<A> {
        a
    }

    fn xsbind<AB, A, B, E>(
        ab: Self::ExtensionSource<AB>,
        t2ab: impl FnOnce(AB) -> (A, B),
        ce: impl FnOnce(Self::ExtensionSource<B>) -> Self::ExtensionResult<B, E>,
        ab2t: impl FnOnce(A, B) -> Result<AB, E>,
    ) -> Self::ExtensionResult<AB, E> {
        let (b, c) = t2ab(ab);
        ab2t(b, ce(c)?)
    }
}
